from logging import info
from threading import Event
from time import sleep
from typing import Optional

from mail_office.models import EmailMessage
from mail_office.postman import Postman
from mail_office.queue_ import Queue


class MailOffice:
    """
    This class is similar to a post office from the real world.

    The principle of operation is as follows:
    the user of the application creates an instance of the email message model,
    which, after creation, is located in the database. MailOffice retrieves from
    the database all instances created by the user and puts them in a queue,
    from which they are received and sent by postmen.
    """

    def __init__(
        self,
        min_queue_size: int,
        max_queue_size: int,
        delete_sent_messages: bool,
        new_messages_wait_time: int
    ) -> None:
        self.shutdown_event = Event()
        self.postman_threads: list[Postman] = []
        self.delete_sent_messages = delete_sent_messages

        self.queue = Queue()
        self.min_queue_size = min_queue_size
        self.max_queue_size = max_queue_size
        self.new_messages_wait_time = new_messages_wait_time

    def push_email_messages(self) -> Optional[int]:
        """
        Queues as many instances of the email message model as needed to
        reach the maximum number of objects in the queue (max_queue_size).

        Filling does not always happen. To retrieve objects from the database
        less often, filling begins when the number of messages in the queue at
        the time of the call is minimal (min_queue_size).

        The queue is constantly filled with objects, if the user has specified
        "min_queue_size = 0".
        """
        if (current_messages_count := self.queue.qsize) <= self.min_queue_size:
            quantity_to_fill = self.max_queue_size - current_messages_count

            if email_messages := EmailMessage.objects.choose_to_send(quantity_to_fill):
                for email_message in email_messages:
                    print("putted email message")
                    self.queue.put(email_message)

                return len(email_messages)

            sleep(self.new_messages_wait_time)

    def start_mailing(self, postmen_count: int) -> None:
        """
        Launches postmen and an endless loop, each iteration of
        which includes replenishing the queue with email messages.
        """
        self.start_postmen(postmen_count)

        while True:
            if pushed_count := self.push_email_messages():
                info(
                    "%d email messages pushed to the queue.",
                    pushed_count
                )

    def start_postmen(self, count: int) -> None:
        """
        Creates and launches postmen. In addition,
        adds them to the list so that later they can be stopped.
        """
        for _ in range(count):
            postman_thread = Postman(self.queue,
                                     self.shutdown_event,
                                     self.delete_sent_messages)
            postman_thread.start()
            self.postman_threads.append(postman_thread)

    def stop_postmen(self) -> None:
        """
        Signals through threading.Event
        to all postmen that it is time to exit.

        The reason why threading.Event is used is documented
        in the run method of the Postman.
        """
        self.shutdown_event.set()

        for postman_thread in self.postman_threads:
            postman_thread.join()


from django.core.management.base import BaseCommand

from mail_office.mail_office import MailOffice
from mail_office.models import EmailMessage
from mail_office.settings import SETTINGS


class Command(BaseCommand):
    def handle(self, *args, **kwargs) -> None:
        postmen_count = SETTINGS.pop("POSTMEN_COUNT")

        mail_office = MailOffice(
            **{
                setting_name.lower(): setting_value
                for setting_name, setting_value in SETTINGS.items()
            }
        )

        EmailMessage.objects.reset_messages()

        self.stdout.write("Starting mailing.")
        try:
            mail_office.start_mailing(postmen_count)
        except (KeyboardInterrupt, SystemExit):
            self.stdout.write(
                "Received KeyboardInterrupt\SystemExit."
                "Completion of mailing."
            )
            mail_office.stop_postmen()


from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.postgres.fields import ArrayField
from django.core.mail import EmailMessage as EmailMessage_
from django.db import models
from django.db.transaction import atomic
from django.utils.module_loading import import_string
from django.utils.timezone import now

if TYPE_CHECKING:
    from django.template.backends.base import BaseEngine


template_backend_path = settings.TEMPLATES[0]["BACKEND"]
template_backend: "BaseEngine" = import_string(template_backend_path)


get_rendered_template_string = (
    lambda template_string, context:
    template_backend.from_string(template_string).render(context)
)


class EmailMessageManager(models.Manager):
    @atomic
    def reset_messages(self) -> None:
        """
        Allows to next start not ignore messages that have been selected
        for sending, but not sent due to the completion of the application.
        """
        self.select_for_update() \
            .filter(status=EmailMessage.Statuses.in_sending) \
            .update(status=EmailMessage.Statuses.new)

    @atomic
    def choose_to_send(self, count: int) -> models.QuerySet["EmailMessage"]:
        base_queryset = self.select_for_update().filter(
            status=EmailMessage.Statuses.new, send_time=now()
        )
        pks = list(
            base_queryset[:count].values_list('pk', flat=True)
        )
        messages_to_send = self.select_for_update().filter(id__in=pks)
        messages_to_send.update(status=EmailMessage.Statuses.in_sending)
        return messages_to_send


class EmailMessage(models.Model):
    class Statuses(models.TextChoices):
        new = "new"
        in_sending = "in_sending"
        canceled = "canceled"
        failed = "failed"
        sent = "sent"

    class Priorities(models.IntegerChoices):
        high = 1
        medium = 2
        low = 3

    status = models.CharField(
        max_length=10,
        choices=Statuses.choices,
        default=Statuses.new
    )

    priority = models.PositiveSmallIntegerField(
        choices=Priorities.choices,
        default=Priorities.high
    )

    subject = models.CharField(max_length=998, blank=True)

    plain_body = models.TextField(blank=True)

    fail_silently = models.BooleanField(default=False)

    to = ArrayField(models.EmailField(blank=True), blank=True)

    send_time = models.DateTimeField(default=now)

    objects = EmailMessageManager()

    class Meta:
        ordering = ("priority")

    # def get_django_email_message(self):
    #     kwargs = {
    #         "from_email": self.from_email,
    #         "to": self.to,
    #         "bcc": self.bcc,
    #         "headers": self.headers,
    #         "cc": self.cc,
    #         "reply_to": self.reply_to,
    #
    #         "subject": get_rendered_template_string(
    #             self.subject, self.context
    #         ),
    #         "body": get_rendered_template_string(
    #             self.plain_body, self.context
    #         )
    #     }
    #
    #     if html_body := get_rendered_template_string(
    #         self.html_body,
    #         self.context
    #     ):
    #         email_message = EmailMultiAlternatives(**kwargs)
    #         email_message.attach_alternative(html_body, "text/html")
    #     else:
    #         return EmailMessage_(**kwargs)
    #
    #     return email_message

    def send(self) -> None:
        print("sending email message")
        email_message = EmailMessage_(subject=self.subject, body=self.plain_body, to=self.to)
        email_message.send(self.fail_silently)

    def process_sent(self, need_to_delete: bool) -> None:
        if need_to_delete:
            print("message sent.deleting.")
            self.delete()
            return

        self.status = self.Statuses.sent
        self.save(update_fields=["status"])

    def process_failed(self, reason: Exception) -> None:
        pass



from queue import Empty
from threading import Thread
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from threading import Event
    from mail_office.models import EmailMessage
    from mail_office.queue_ import Queue


class Postman(Thread):
    """
    Sender of queued messages (Worker).

    This is the only class in the application that interacts
    with email message instances.

    In fact, the job of this class is to use instance methods in the background.
    """

    def __init__(
        self,
        queue: "Queue",
        shutdown_event: "Event",
        delete_sent_messages: bool
    ) -> None:
        super().__init__()
        self.queue = queue
        self.shutdown_event = shutdown_event
        self.delete_sent_messages = delete_sent_messages

    def run(self) -> None:
        """
        Starts a cycle for sending messages.

        Note that the postman exits based on threading.Event and not
        on any magic value received from the queue.

        Firstly, this approach eliminates the need to clear the queue
        (in order to prioritize the magic value) and queue this magic
        value in the number corresponding to the number of postmen.

        Secondly, if the thread is made daemonic, then if the postman
        took a message from the queue and the program termination was caused,
        the postman will not have time to send the message, which will lead
        to the need to send this message at the next start.
        """
        while not self.shutdown_event.is_set():
            try:
                email_message: "EmailMessage" = self.queue.get(timeout=2)
            except Empty:
                # It is perceived as the fact that no new messages have
                # appeared in the queue at the time "get" is called.
                continue

            self.send_message(email_message)

    def send_message(self, email_message: "EmailMessage") -> None:
        try:
            email_message.send()
            email_message.process_sent(self.delete_sent_messages)
        except Exception as exc:
            print("send failed")
            email_message.process_failed(exc)



# Generated by Django 3.2 on 2021-04-20 13:55

import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='EmailMessageTemplate',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subject', models.CharField(blank=True, max_length=998)),
                ('plain_body', models.TextField(blank=True)),
                ('html_body', models.TextField(blank=True)),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EmailMessage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subject', models.CharField(blank=True, max_length=998)),
                ('plain_body', models.TextField(blank=True)),
                ('html_body', models.TextField(blank=True)),
                ('status', models.CharField(choices=[('new', 'New'), ('in_sending', 'In Sending'), ('canceled', 'Canceled'), ('failed', 'Failed'), ('sent', 'Sent')], default='new', editable=False, max_length=10)),
                ('priority', models.PositiveSmallIntegerField(choices=[(1, 'High'), (2, 'Medium'), (3, 'Low')], default=1)),
                ('name', models.CharField(max_length=100)),
                ('retries_number', models.PositiveSmallIntegerField(default=0)),
                ('max_retries_number', models.PositiveSmallIntegerField(default=1)),
                ('retry_delay', models.PositiveIntegerField(default=900)),
                ('from_email', models.EmailField(blank=True, max_length=254)),
                ('fail_silently', models.BooleanField(default=False)),
                ('reply_to', django.contrib.postgres.fields.ArrayField(base_field=models.EmailField(blank=True, max_length=254), blank=True, size=None)),
                ('to', django.contrib.postgres.fields.ArrayField(base_field=models.EmailField(blank=True, max_length=254), blank=True, size=None)),
                ('cc', django.contrib.postgres.fields.ArrayField(base_field=models.EmailField(blank=True, max_length=254), blank=True, size=None)),
                ('bcc', django.contrib.postgres.fields.ArrayField(base_field=models.EmailField(blank=True, max_length=254), blank=True, size=None)),
                ('headers', models.JSONField(blank=True)),
                ('context', models.JSONField(blank=True)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('send_time', models.DateTimeField(default=django.utils.timezone.now)),
                ('template', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='mail_office.emailmessagetemplate')),
            ],
            options={
                'ordering': ('priority', 'created_at'),
            },
        ),
    ]



# Generated by Django 3.2 on 2021-04-20 17:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mail_office', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='emailmessage',
            name='template',
        ),
        migrations.AlterField(
            model_name='emailmessage',
            name='status',
            field=models.CharField(choices=[('new', 'New'), ('in_sending', 'In Sending'), ('canceled', 'Canceled'), ('failed', 'Failed'), ('sent', 'Sent')], default='new', max_length=10),
        ),
        migrations.DeleteModel(
            name='EmailMessageTemplate',
        ),
    ]



from django.conf import settings


SETTINGS = {
    "POSTMEN_COUNT": 2,
    "DELETE_SENT_MESSAGES": False,

    "MIN_QUEUE_SIZE": 5,
    "MAX_QUEUE_SIZE": 20,
    "NEW_MESSAGES_WAIT_TIME": 2
}

SETTINGS.update(
    getattr(
        settings, "MAIL_OFFICE", {}
    )
)